import os

import random
from pathlib import Path

from cocotb.triggers import Timer
from cocotb.runner import get_runner, get_results

import siliconcompiler
from siliconcompiler.tools.slang import elaborate


def run_cocotb(
        chip: siliconcompiler.Chip,
        test_module_name,
        output_dir_name=None,
        simulator_name="icarus",
        timescale=None,
        parameters=None):

    # Use surelog to pickle Verilog sources
    flow = "cocotb_flow"
    chip.node(flow, "import", elaborate)
    chip.set("option", "flow", flow)
    assert chip.run()

    pickled_verilog = chip.find_result("v", "import")
    assert pickled_verilog, "Could not locate pickled verilog"

    if output_dir_name is None:
        output_dir_name = test_module_name

    top_level_dir = os.getcwd()
    build_dir = Path(chip.getbuilddir()) / output_dir_name
    test_dir = None

    results_xml = None
    # Need to check if we are running inside of pytest. See link below.
    # https://github.com/cocotb/cocotb/blob/d883ce914063c3601455d10a40f459fffa22d8f2/cocotb/runner.py#L313
    if not os.getenv("PYTEST_CURRENT_TEST", None):
        results_xml = build_dir / "results.xml"
        test_dir = top_level_dir

    # Build HDL in chosen simulator
    runner = get_runner(simulator_name)
    runner.build(
        sources=[pickled_verilog],
        hdl_toplevel=chip.design,
        waves=True,
        timescale=timescale,
        build_dir=build_dir,
        parameters=parameters
    )
    # Run test
    _, tests_failed = get_results(runner.test(
        hdl_toplevel=chip.top(),
        test_module=test_module_name,
        test_dir=test_dir,
        results_xml=results_xml,
        build_dir=build_dir,
        timescale=timescale,
        waves=True
    ))

    return tests_failed


async def drive_reset(reset, reset_time_in_ns=100, active_level=False):
    reset.value = 1 if active_level else 0
    await Timer(reset_time_in_ns, units="ns")
    reset.value = 0 if active_level else 1


def random_bool_generator():
    while True:
        yield (random.randint(0, 1) == 1)



import random
from decimal import Decimal

import siliconcompiler

import cocotb
from cocotb.clock import Clock
from cocotb.triggers import ClockCycles, Timer, Combine
from cocotb.regression import TestFactory
from cocotb import utils

from lambdalib import ramlib
from lambdalib.utils._tb_common import (
    run_cocotb,
    drive_reset,
    random_bool_generator
)
from lambdalib.ramlib.tests.la_asyncfifo import (
    LaAsyncFifoWrBus,
    LaAsyncFifoRdBus,
    LaAsyncFifoSource,
    LaAsyncFifoSink
)


def bursty_en_gen(burst_len=20):
    while True:
        en_state = (random.randint(0, 1) == 1)
        for _ in range(0, burst_len):
            yield en_state


@cocotb.test()
async def test_almost_full(dut):

    wr_clk_period_ns = 10.0
    rd_clk_period_ns = 10.0

    fifo_source = LaAsyncFifoSource(
        bus=LaAsyncFifoWrBus.from_prefix(dut, ""),
        clock=dut.wr_clk,
        reset=dut.wr_nreset
    )

    fifo_sink = LaAsyncFifoSink(
        bus=LaAsyncFifoRdBus.from_prefix(dut, ""),
        clock=dut.rd_clk,
        reset=dut.rd_nreset
    )
    fifo_sink.pause()

    # Reset DUT
    await Combine(
        cocotb.start_soon(drive_reset(dut.wr_nreset)),
        cocotb.start_soon(drive_reset(dut.rd_nreset))
    )

    await cocotb.start(Clock(dut.wr_clk, wr_clk_period_ns, units="ns").start())
    # Randomize phase shift between clocks
    await Timer(wr_clk_period_ns * random.random(), "ns", round_mode="round")
    await cocotb.start(Clock(dut.rd_clk, rd_clk_period_ns, units="ns").start())

    almost_full_level = int(dut.AFULLFINAL.value)

    await ClockCycles(dut.wr_clk, 3)

    # Almost full should be low before writing to FIFO
    assert dut.wr_almost_full.value == 0

    # Write to FIFO
    for i in range(0, almost_full_level):
        await fifo_source.send(i)

    # Wait for write to complete
    await fifo_source.wait_until_idle()
    assert dut.wr_almost_full.value == 1

    # Read one value out of FIFO
    fifo_sink.resume()
    await fifo_sink.read()
    fifo_sink.pause()

    # Check that almost full is lowered
    await ClockCycles(dut.wr_clk, 4)
    assert dut.wr_almost_full.value == 0


async def fifo_rd_wr_test(
    dut,
    wr_clk_period_ns=10.0,
    rd_clk_period_ns=10.0,
    wr_en_generator=None,
    rd_en_generator=None
):

    fifo_source = LaAsyncFifoSource(
        bus=LaAsyncFifoWrBus.from_prefix(dut, ""),
        clock=dut.wr_clk,
        reset=dut.wr_nreset
    )
    fifo_source.set_wr_en_generator(random_bool_generator())

    fifo_sink = LaAsyncFifoSink(
        bus=LaAsyncFifoRdBus.from_prefix(dut, ""),
        clock=dut.rd_clk,
        reset=dut.rd_nreset
    )
    fifo_sink.set_rd_en_generator(random_bool_generator())

    # Reset DUT
    await Combine(
        cocotb.start_soon(drive_reset(dut.wr_nreset)),
        cocotb.start_soon(drive_reset(dut.rd_nreset))
    )

    await cocotb.start(Clock(dut.wr_clk, wr_clk_period_ns, units="ns").start())
    # Randomize phase shift between clocks
    await Timer(wr_clk_period_ns * random.random(), "ns", round_mode="round")
    await cocotb.start(Clock(dut.rd_clk, rd_clk_period_ns, units="ns").start())

    await ClockCycles(dut.wr_clk, 3)

    expected_len = 256
    expected = [random.getrandbits(len(dut.wr_din)) for _ in range(expected_len)]

    for val in expected:
        await fifo_source.send(val)

    actual = [await fifo_sink.read() for _ in range(expected_len)]

    assert actual == expected

    await ClockCycles(dut.wr_clk, 10)


# Generate sets of tests based on the different permutations of the possible arguments to fifo_test
MAX_PERIOD_NS = 10.0
MIN_PERIOD_NS = 1.0
# Generate random clk period to test between min and max
RAND_WR_CLK_PERIOD_NS, RAND_RD_CLK_PERIOD_NS = [utils.get_time_from_sim_steps(
    # Time step must be even for cocotb clock driver
    steps=utils.get_sim_steps(
        time=Decimal(MIN_PERIOD_NS) + (
            Decimal(MAX_PERIOD_NS - MIN_PERIOD_NS)
            * Decimal(random.random()).quantize(Decimal("0.00"))
        ),
        units="ns",
        round_mode="round"
    ) & ~1,
    units="ns"
) for _ in range(0, 2)]

# Factory to automatically generate a set of tests based on the different permutations
# of the provided test arguments
tf = TestFactory(fifo_rd_wr_test)
tf.add_option('wr_clk_period_ns', [MIN_PERIOD_NS, RAND_WR_CLK_PERIOD_NS, MAX_PERIOD_NS])
tf.add_option('rd_clk_period_ns', [MIN_PERIOD_NS, RAND_RD_CLK_PERIOD_NS, MAX_PERIOD_NS])
tf.add_option('wr_en_generator', [None, random_bool_generator, bursty_en_gen])
tf.add_option('rd_en_generator', [None, random_bool_generator, bursty_en_gen])
tf.generate_tests()


def test_la_asyncfifo():
    chip = siliconcompiler.Chip("la_asyncfifo")

    chip.input("ramlib/rtl/la_asyncfifo.v", package='lambdalib')
    chip.use(ramlib)

    for depth in [2, 4, 8]:
        test_module_name = "lambdalib.ramlib.tests.tb_la_asyncfifo"
        test_name = f"{test_module_name}_depth_{depth}"
        tests_failed = run_cocotb(
            chip=chip,
            test_module_name=test_module_name,
            timescale=("1ns", "1ps"),
            parameters={
                "DW": 32,
                "DEPTH": depth
            },
            output_dir_name=test_name
        )
        assert (tests_failed == 0), f"Error test {test_name} failed!"


if __name__ == "__main__":
    test_la_asyncfifo()



import cocotb
from cocotb.triggers import RisingEdge
from cocotb_bus.bus import Bus
from cocotb.queue import Queue
from cocotb.binary import BinaryValue


class LaAsyncFifoWrBus(Bus):
    """Cocotb bus for lambdalib async FIFO WR interface"""

    _signals = ["wr_din", "wr_en", "wr_full"]
    _optional_signals = ["wr_almost_full", "wr_chaosmode"]

    def __init__(self, entity=None, prefix=None, **kwargs):
        super().__init__(
            entity,
            prefix,
            self._signals,
            optional_signals=self._optional_signals,
            **kwargs
        )

    @classmethod
    def from_entity(cls, entity, **kwargs):
        return cls(entity, **kwargs)

    @classmethod
    def from_prefix(cls, entity, prefix, **kwargs):
        return cls(entity, prefix, **kwargs)


class LaAsyncFifoRdBus(Bus):
    """Cocotb bus for lambdalib async FIFO RD interface"""

    _signals = ["rd_dout", "rd_en", "rd_empty"]
    _optional_signals = []

    def __init__(self, entity=None, prefix=None, **kwargs):
        super().__init__(
            entity,
            prefix,
            self._signals,
            optional_signals=self._optional_signals,
            **kwargs
        )

    @classmethod
    def from_entity(cls, entity, **kwargs):
        return cls(entity, **kwargs)

    @classmethod
    def from_prefix(cls, entity, prefix, **kwargs):
        return cls(entity, prefix, **kwargs)


class LaAsyncFifoSource:
    """Driver for write side of lambdalib async fifo"""

    def __init__(self, bus: LaAsyncFifoWrBus, clock, reset=None):
        self.bus = bus
        self.clock = clock
        self.reset = reset

        self.queue = Queue()
        self.width = len(self.bus.wr_din)

        self._run_cr = cocotb.start_soon(self._run())

        self.bus.wr_chaosmode.value = 0
        self.bus.wr_en.value = 0

        self.wr_en_generator = None

    def set_wr_en_generator(self, generator):
        self.wr_en_generator = generator

    async def send(self, data):
        await self.queue.put(data)

    async def wait_until_idle(self):
        while not self.queue.empty() or self.bus.wr_en.value == 1:
            await RisingEdge(self.clock)

    async def _run(self):
        clock_edge_event = RisingEdge(self.clock)

        bus_val = None

        while True:
            await clock_edge_event

            fifo_full = self.bus.wr_full.value
            wr_en = self.bus.wr_en.value

            if bus_val is None:
                bus_val = await self.queue.get()
            elif wr_en and not fifo_full:
                bus_val = None if self.queue.empty() else self.queue.get_nowait()

            if bus_val is None:
                self.bus.wr_en.value = 0
            else:
                self.bus.wr_en.value = next(self.wr_en_generator) if self.wr_en_generator else 1

            self.bus.wr_din.value = bus_val if bus_val else 0


class LaAsyncFifoSink:
    """Driver for read side of lambdalib async fifo"""

    def __init__(self, bus: LaAsyncFifoRdBus, clock, reset=None):
        self.bus = bus
        self.clock = clock
        self.reset = reset

        self.queue = Queue()
        self.width = len(self.bus.rd_dout)

        self._run_cr = cocotb.start_soon(self._run())

        self.bus.rd_en.value = 0

        self.rd_en_generator = None
        self._pause = False

    def set_rd_en_generator(self, generator):
        self.rd_en_generator = generator

    def pause(self):
        self.bus.rd_en.value = 0
        self._pause = True

    def resume(self):
        self._pause = False

    async def read(self) -> BinaryValue:
        return await self.queue.get()

    async def _run(self):
        clock_edge_event = RisingEdge(self.clock)

        while True:
            await clock_edge_event

            fifo_empty = self.bus.rd_empty.value

            if self._pause:
                self.bus.rd_en.value = 0
            elif self.rd_en_generator:
                self.bus.rd_en.value = next(self.rd_en_generator)
            else:
                self.bus.rd_en.value = 1

            if self.bus.rd_en.value and not fifo_empty:
                self.queue.put_nowait(self.bus.rd_dout.value)
